<?php

namespace ApplicationsModule\MailModule;

use \dibi;
use DibiException;
use Celebrio\Core\ServiceProvider;
use Celebrio\Modules;
use \Nette\Environment;
use KernelModule\SignModel;
use Celebrio\ShadowReader;

class StoredMailManagement {
	const FROM = "from";
	const TO = "to";
	const REPLY = "reply";
	const CC = "cc";
	const BCC = "bcc";

    private static $sendgridUserLogin = null;
    private static $sendgridUserPass = null;
    private $server;
    private $service;

    public function __construct($server = null) {
        $this->server = $server;
    }

    /**
     * Process mail from sendgrid
     * @param array $data - $_POST
     * @param array $files - $_FILES
     * @return false if failed
     */
    public function processMail($data, $files = null) {

        $mail = new Mail();



        $envelope = json_decode($data["envelope"]);
        $server = $this->getServerIdByAddress($envelope->to[0]);
        $this->server = $server;





        if (isset($data["subject"])) {
            $mail->setSubject($data["subject"]);
        }



        if (isset($data["from"])) {
            $mail->setFrom($this->processAddresses($data["from"]));
        }

        if (isset($data["to"])) {
            $mail->setToList($this->processAddresses($data["to"]));
        }

        if (isset($data["cc"])) {
            $mail->setCcList($this->processAddresses($data["cc"]));
        }

        if (isset($data["bcc"])) {
            $mail->setBccList($this->processAddresses($data["bcc"]));
        }


        if (isset($data["bcc"])) {
            $mail->setBccList($this->processAddresses($data["bcc"]));
        }



        if (isset($data["headers"])) {
            $mail->setHeaders($data["headers"]);
        }



        if (isset($data["text"]) && $data["text"] != "") {
            $mail->setBody($data["text"]);
        }
        if (isset($data["html"]) && $data["html"] != "") {
            $mail->setBody($data["html"]);
        }

        $mail->setStatus("INBOX");
        $mail->setUnread(true);

        if (!empty($files)) {
            $mail->setAttachment(true);
        } else {
            $mail->setAttachment(false);
        }


        $mail->setComplete(true);
        if (isset($files)) {
            $mail->setAttachmentTempFiles($files);
        }


        //log user sendgrid to permission to save files
        $this->logAsSendGrid();

        return $this->addMailToDB($mail);
    }

    /**
     * Process addresses from string
     * @param string $addresses
     * @return array Address
     */
    public function processAddresses($addresses) {
        if($addresses == null) return null;
        
        $finnalAddresses = array();
        $addressesArray = explode(", ", $addresses);
        foreach ($addressesArray as $addr) {

            $mainAddr = preg_replace("/(.*)(<)(.*)(>)(.*)/", "$3", $addr);
            if (preg_match("/(.*)( <)(.*)(>)(.*)/", $addr)) {
                $name = preg_replace("/(.*)( <)(.*)(>)(.*)/", "$1", $addr);
                    if (preg_match("/(\")(.*)(\")/", $name)) {
                        $name = preg_replace("/(\")(.*)(\")/", "$2", $name);
                    }
            } else {
                $name = null;
            }



            $address = new Address();
            $address->setAddress($mainAddr);
            $address->setName($name);

            $this->addAddressToDB($address);
            $finnalAddresses[] = $address;
        }

        if (empty($finnalAddresses)) {
            return null;
        }
        return $finnalAddresses;
    }

    /**
     * Adding address to DB or update existing if name is different
     * @param Address $address
     * @return false if failed
     */
    public function addAddressToDB(Address $address) {
        try {

            $exists = dibi::select("*")->from("addresses")->where("address = %s", $address->getAddress())->fetch();


            if (!$exists) {

                dibi::insert("addresses", array("address" => $address->getAddress(), "name" => $address->getName()))->execute();

                $address->setId(dibi::getInsertId());
            } else {
                $address->setId($exists["id_address"]);
                if ($address->getName() != null && $address->getName() != $exists["name"]) {
                    dibi::update("addresses", array("name" => $address->getName()))->where("id_address = %i", $address->getId())->execute();
                }
            }
        } catch (\DibiException $e) {
            \Logger::getRootLogger()->warn("Database problem while storing address.", $e);
            return false;
        }
        return $address;
    }

    /**
     * Update uncomplete mail and make it complete + store mail on local or azure
     * @param Mail $mail
     * @return false if failed
     */
    public function updateCompleteMailToDB(Mail $mail) {
        /*
        if(strlen ($mail->getBody()) > 64000) {
            return false;
        }
        if(strlen ($mail->getHeaders()) > 64000) {
            return false;
        }*/
        
        try {
            dibi::begin();
            $update = array("complete" => $mail->getComplete(),
                "attachment" => $mail->getAttachment());

            dibi::update("mailbox", $update)->where("id_mail = %i", $mail->getId())->execute();

            $this->addRelationsToDB($mail, self::TO);
            $this->addRelationsToDB($mail, self::REPLY);
            $this->addRelationsToDB($mail, self::CC);
            $this->addRelationsToDB($mail, self::BCC);

            $this->storeMailContent($mail);


            dibi::commit();
        } catch (\DibiException $e) {
            \Logger::getRootLogger()->warn("Database problem while updating mail.", $e);
            return false;
        }
    }

    /**
     * Adding mail to DB, if mail has set complete as true, save content of mail on local or azure
     * @param Mail $mail
     * @return false if failed
     */
    public function addMailToDB(Mail $mail) {
        if ($this->server != null) {
            if ($mail->getTime() == null) {
                $time = new \DateTime();
                $time->setTimestamp(time());
                $mail->setTime($time);
            }
            try {
                $insert = array("subject" => $mail->getSubject(),
                    "time" => $mail->getTime(),
                    "unread" => false,
                    "status" => "INBOX",
                    "server" => $this->server,
                    "message_uid" => $mail->getMailUID(),
                    "complete" => $mail->getComplete(),
                    "attachment" => $mail->getAttachment());


                dibi::begin();
                dibi::insert("mailbox", $insert)->execute();


                $mail->setId(dibi::getInsertId());


                $this->addRelationsToDB($mail, self::FROM);

                $this->addRelationsToDB($mail, self::TO);
                $this->addRelationsToDB($mail, self::REPLY);
                $this->addRelationsToDB($mail, self::CC);
                $this->addRelationsToDB($mail, self::BCC);

                if ($mail->getComplete()) {
                    $this->storeMailContent($mail);
                }

                if ($mail->getAttachmentTempFiles() != null) {
                    $acd = new AttachCommonDriver();
                    $acd->saveAttachments($mail);
                }

                dibi::commit();
            } catch (\DibiException $e) {
                \Logger::getRootLogger()->warn("Database problem while storing mail.", $e);
                return false;
            }
            
        } else {
            return false;
        }
        
        return true;
    }

    /**
     * processing mail relations and adding to DB
     * @param Mail $mail, $type
     */
    public function addRelationsToDB(Mail $mail, $type) {
        $addresses = array();

        switch ($type) {
            case self::FROM:
                $addresses = $mail->getFrom();
                break;
            case self::TO:
                $addresses = $mail->getToList();
                break;
            case self::REPLY:
                $addresses = $mail->getReply();
                break;
            case self::CC:
                $addresses = $mail->getCcList();
                break;
            case self::BCC:
                $addresses = $mail->getBccList();
                break;
            default:
                \Logger::getRootLogger()->error("Unknown mail address realtionship.");
                break;
        }


        if ($addresses != null) {
            foreach ($addresses as $addr) {
                $relation = Array();
                $relation["mail"] = $mail;
                $relation["address"] = $addr;
                $relation["type"] = $type;

                $this->addRelationToDB($relation);
            }
        }
    }

    /**
     * Adding mail relation to DB
     * @param array $relation
     * @return false if failed
     */
    public function addRelationToDB($relation) {

        try {
            $insert = array("id_mail" => $relation["mail"]->getId(),
                            "id_address" => $relation["address"]->getId(),
                            "type" => $relation["type"]);


            dibi::insert("addresses_mailbox", $insert)->execute();
        } catch (\DibiException $e) {
            \Logger::getRootLogger()->warn("Database problem while storing address relation.", $e);
            return false;
        }
    }

    /**
     * Getting uncomplete mails headers from DB
     * @param string $complete - self::ALL, self::COMPLETE, self::INCOMPLETE
     * @return null if failed
     */
    public function getUncompleteMails() {
        if ($this->server != null) {

            $mails = null;
            try {
                $query = dibi::select("*")->from("mailbox")->where("server = %i", $this->server);
                $result = $query->where("complete = %i", false)->fetchAll(0, 20);
        

                foreach ($result as $row) {
                    $mail = $this->buildMail($row);

                    $mails[] = $mail;
                }
                if(empty($mails)) {
                    return false;
                }
                return $mails;
            } catch (\DibiException $e) {
                \Logger::getRootLogger()->warn("Database problem while getting mails.", $e);
                return false;
            }
        }
    }

    /**
     * 
     * 
     * @param array $servers (id => server)
     * @param int $from range from
     * @param int $to ragne to
     * @return array Mail
     */
    public function getMailsByServersInRange($servers, $from, $to) {
        $query = dibi::select("*")->from("mailbox");

        $serversId = array();

        foreach ($servers as $key => $value) {
            $serversId[] = $key;
        }

        $query->where(array("server%in" => $serversId))
                ->orderBy("time")->desc();
        
        //TODO: hack for sqlserver :-/
        $driver = dibi::getConnection()->getConfig("driver");
        if ( $driver != "mssql2005" && $driver != "azure") {
            $result = $query->fetchAll($from, $to - $from);
        } else {
            $q = dibi::query("SELECT * FROM (
            SELECT TOP ".($to-$from)." * FROM (
            SELECT TOP ".($from+($to-$from))." * FROM mailbox
            WHERE server IN %in
            ORDER BY time DESC) q ORDER BY time ASC) t ORDER BY time DESC", $serversId);
            $result = $q->fetchAll();
        }
        
        
        $mails = array();

        foreach ($result as $row) {
            $mails[] = $this->buildMail($row);
        }

        return $mails;
    }

    /**
     * getting last message UID of stored mails
     * @return last uid or null if failed
     */
    function getLastUID() {
        if ($this->server != null) {
            try {
                $result = dibi::select("MAX(message_uid) as max")-> from("mailbox")->where("server = %i", $this->server)->fetch();

                
                
                if ($result["max"] == null) {
                    return 0;
                } else {
                    return $result["max"];
                }
            } catch (\DibiException $e) {

                \Logger::getRootLogger()->warn("Database problem while getting last mail UID.", $e);
                return null;
            }
        }
    }

    /**
     * Getting whole mail from DB and local or azure
     * @param int $idMail
     * @return false if failed
     */
    function getMail($idMail) {
        if ($this->server != null) {

            try {
                $result = dibi::select("*")->from("mailbox")->where("server = %i", $this->server)->where("id_mail = %i", $idMail)->fetch();
                $mail = $this->buildMail($result);

                if ($mail->getComplete()) {
                    $this->getMailContent($mail);
                }


                return $mail;
            } catch (\DibiException $e) {

                \Logger::getRootLogger()->warn("Database problem while getting mail.", $e);
                return false;
            }
        }
    }

    /**
     * Assemble mail from fetch row
     * @param array $row
     * @return Mail
     */
    private function buildMail($row) {
        $mail = new Mail();
        $mail->setId($row["id_mail"]);
        $mail->setSubject($row["subject"]);
        $mail->setTime($row["time"]);
        $mail->setUnread($row["unread"]);
        $mail->setStatus($row["status"]);
        $mail->setServer($row["server"]);
        $mail->setMailUID($row["message_uid"]);

        $mail->setAttachment($row["attachment"]);
        $mail->setComplete($row["complete"]);

        $this->buildRelationsToMail($mail);

        return $mail;
    }

    /**
     * getting relations and set it to mail
     * @param Mail $mail
     * @return false if failed
     */
    public function buildRelationsToMail(Mail $mail) {
        try {
            $from = null;
            $to = null;
            $reply = null;
            $cc = null;
            $bcc = null;

            $result = dibi::select("*")
                    ->from("addresses_mailbox")
                    ->innerJoin("addresses")
                    ->on("addresses_mailbox.id_address = addresses.id_address")
                    ->where("addresses_mailbox.id_mail = %i", $mail->getId())
                    ->fetchAll();

            foreach ($result as $row) {
                $address = $this->buildAddress($row);
                switch ($row["type"]) {
                    case self::FROM:
                        $from = $address;
                        break;
                    case self::TO:
                        $to[] = $address;
                        break;
                    case self::REPLY:
                        $reply[] = $address;
                        break;
                    case self::CC:
                        $cc[] = $address;
                        break;
                    case self::BCC:
                        $bcc[] = $address;
                        break;
                    default:
                        throw new \Exception();
                        \Logger::getRootLogger()->error("Unknown mail address realtionship.");
                        break;
                }
            }

            $mail->setFrom($from);
            $mail->setToList($to);
            $mail->setReply($reply);
            $mail->setCcList($cc);
            $mail->setBccList($bcc);
        } catch (\DibiException $e) {
            \Logger::getRootLogger()->warn("Database problem while getting relations.", $e);
            return false;
        }
    }

    /**
     * getting address from DB by id
     * @param Int $id
     * @return Address or false if failed
     */
    public function getAddress($id) {
        try {

            $result = dibi::select("*")->from("addresses")->where("id_address = %i", $id)->fetch();

            return $this->buildAddress($result);
        } catch (\DibiException $e) {
            \Logger::getRootLogger()->warn("Database problem while getting address.", $e);
            return false;
        }
    }

    /**
     * Assemble address from fetch row
     * @param array $row
     * @return Address
     */
    private function buildAddress($row) {
        $address = new Address();
        $address->setId($row["id_address"]);
        $address->setAddress($row["address"]);
        $address->setName($row["name"]);

        return $address;
    }

    /**
     * Delete mail by id
     * @param Int $id
     * @return false if failed
     */
    public function deleteMail($id) {
        try {

            dibi::delete("mailbox")->where("id_mail = %i", $id)->execute();
        } catch (\DibiException $e) {
            \Logger::getRootLogger()->warn("Database problem while deleting mail.", $e);
            return false;
        }

        $this->deleteMailContent($id);
    }

    /**
     * Getting server id by address
     * @param string $address
     * @return Int id or false if failed
     */
    public function getServerIdByAddress($address) {
        try {

            $result = dibi::select("id")->from("celebriomailapp")->where("mailaddress = %s", $address)->fetch();

            return $result["id"];
        } catch (\DibiException $e) {
            \Logger::getRootLogger()->warn("Database problem while getting server id.", $e);
            return false;
        }
    }

    private function storeMailContent(Mail $mail) {
        $service = $this->getService();
        $service->saveMailContent($mail);
    }

    private function getMailContent(Mail $mail) {
        $service = $this->getService();
        $service->loadMailContent($mail);
    }

    private function deleteMailContent($id) {
        $service = $this->getService();
        $service->deleteMailContent($id);
    }

    private function getService() {
        if (!isset($this->service)) {

            /*/
            $this->service = new Azure();
            /*/
            if (AZURE) {
                $this->service = new Azure();
            } else {
                $this->service = new Local();
            }/**/
        }

        return $this->service;
    }

    public function getServer() {
        return $this->server;
    }

    public function setServer($email) {
        $this->email = $email;
    }

    public function createCelebrioMailAccount($userId, $userName) {
        try {

            $insert = array("user_id" => $userId,
                            "mailaddress" => $userName . "@celebrio.cz",
                            "provider" => "celebrio");

            dibi::insert("celebriomailapp", $insert)->execute();
        } catch (\DibiException $e) {
            \Logger::getRootLogger()->warn("Database problem while creating mail account.", $e);
            return false;
        }
    }

    /**
     * Set logged user as user sengrid
     */
    private function logAsSendGrid() {
        $user = Environment::getUser();
        $user->setExpiration('+ 1 minutes', TRUE);

        try {
            $sm = new SignModel();
            $sm->signIn($user, array("username" => $this->getSendgridUserLogin(), "password" => $this->getSendgridUserPass()));
        } catch (AuthenticationException $e) {
            \Logger::getRootLogger()->warn("Problem while sign in sendgrid user", $e);
        }
    }

    /**
     * Set login and password for sendgrid user from config
     * @param string $login
     * @param string $pass
     */
    public static function setSendgridUserLogin($login) {
        self::$sendgridUserLogin = $login;
    }

    private function getSendgridUserLogin() {
        if (self::$sendgridUserLogin === null) {
            throw new InvalidStateException("Sendgrid user has not been set up yet.");
        }
        return self::$sendgridUserLogin;
    }

    private function getSendgridUserPass() {
        $shadowReader = new ShadowReader();
        return $shadowReader->getPassword('sendgriduser');
    }

}